
/*******************************************************************************
*
* File check10.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Solution of the little Dirac equation
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "flags.h"
#include "block.h"
#include "linalg.h"
#include "linsolv.h"
#include "sw_term.h"
#include "dirac.h"
#include "eva.h"
#include "sap_gcr.h"
#include "dfl.h"
#include "global.h"

static int nbs,Nss,nvhs;
static complex **vs,**zvs;
static complex_dble **vds,**zvds,*avd,*cs1,*cs2;


static void alloc_arrays(int Ns)
{
   if (Ns>0)
   {
      vs=amalloc(2*Ns*sizeof(*vs),3);
      vds=amalloc(2*Ns*sizeof(*vds),3);
      avd=amalloc((Ns*Ns+2*Ns)*sizeof(*avd),ALIGN);   
      error((vs==NULL)||(vds==NULL)||(avd==NULL),1,"alloc_arrays [dfl.c]",
            "Unable to allocate auxiliary arrays");

      zvs=vs+Ns;
      zvds=vds+Ns;
      cs1=avd+Ns*Ns;
      cs2=cs1+Ns;
   }
}


static void Zee_dble(complex_dble *v,complex_dble *w)
{
   int Ns,nvh,nmat;
   complex_dble *z,*vm;   
   zop_dble_t *zop;

   zop=zopnohat_dble();
   z=(*zop).zee;
   Ns=Nss;
   nvh=nvhs;
   
   nmat=Ns*Ns;
   vm=v+nvh;

   for (;v<vm;v+=Ns)
   {
      cmat_vec_dble(Ns,z,v,w);      
      z+=nmat;
      w+=Ns;
   }   
}


static void Zhatinv_dble(int nmx,double res,complex_dble *v,complex_dble *w,
                         int *status)
{
   int nvh;

   nvh=nvhs;
   Zee_dble(v,w);
   set_vd2zero(nvh,w+nvh);
   dfl_solve_little(nmx,res,w,status);
   set_vd2zero(nvh,w+nvh);
}


static void sum_vprod(int n)
{
   int i;
   
   if (NPROC>1)
   {
      MPI_Reduce((double*)(cs1),(double*)(cs2),2*n,MPI_DOUBLE,
                 MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast((double*)(cs2),2*n,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
   {
      for (i=0;i<n;i++)
      {
         cs2[i].re=cs1[i].re;
         cs2[i].im=cs1[i].im;
      }
   }
}


static void Qvd(complex_dble *v,complex_dble *w)
{
   int Ns,nvh,i;
   complex_dble z;

   Ns=Nss;
   nvh=nvhs;
   assign_vd2vd(nvh,v,w);
   set_vd2zero(nvh,w+nvh);

   for (i=0;i<Ns;i++)
      cs1[i]=vprod_dble(nvh,0,vds[i],w);

   sum_vprod(Ns);
   
   for (i=0;i<Ns;i++)
   {
      z.re=-cs2[i].re;
      z.im=-cs2[i].im;
      mulc_vadd_dble(nvh,w,vds[i],z);
   }
}


static void Lvd(complex_dble *v,complex_dble *w)
{
   int Ns,nvh,i;
   complex_dble z;

   Ns=Nss;
   nvh=nvhs;

   assign_vd2vd(nvh,v,w);
   set_vd2zero(nvh,w+nvh);

   for (i=0;i<Ns;i++)
      cs1[i]=vprod_dble(nvh,0,vds[i],w);

   sum_vprod(Ns);
   cmat_vec_dble(Ns,avd,cs2,cs1);

   for (i=0;i<Ns;i++)
   {
      z.re=-cs1[i].re;
      z.im=-cs1[i].im;      
      mulc_vadd_dble(nvh,w,zvds[i],z);
   }
}


static double lambda0(int nit)
{
   int Ns,nvh;
   int i,iw0,iw1;
   double r;
   complex_dble **wvd;

   Ns=Nss;
   nvh=nvhs;
   
   wvd=reserve_wvd(2);
   iw0=0;
   iw1=1;
   r=0.0;
   random_vd(nvh,wvd[iw0],1.0);
   vnormalize_dble(nvh,1,wvd[iw0]);

   for (i=0;i<nit;i++)
   {
      Zhat_dble(wvd[iw0],wvd[iw1]);
      r=vnormalize_dble(nvh,1,wvd[iw1]);

      iw0=(iw0^0x1);
      iw1=(iw1^0x1);

      message("r=%.2e\n",r);
   }

   release_wvd();
   
   return r;
}


static double lambda1(int nit,int nmx,double res,int *status)
{
   int Ns,nvh,stat;
   int i,iw0,iw1;
   double r;
   complex_dble **wvd;

   Ns=Nss;
   nvh=nvhs;

   wvd=reserve_wvd(2);
   iw0=0;
   iw1=1;
   (*status)=0;
   r=0.0;
   random_vd(nvh,wvd[iw0],1.0);
   vnormalize_dble(nvh,1,wvd[iw0]);

   for (i=0;i<nit;i++)
   {
      Zhatinv_dble(nmx,res,wvd[iw0],wvd[iw1],&stat);
      r=vnormalize_dble(nvh,1,wvd[iw1]);
      
      if ((*status)>=0)
      {
         if (stat>=0)
            (*status)+=stat;
         else
            (*status)=stat;
      }      

      iw0=(iw0^0x1);
      iw1=(iw1^0x1);

      message("r=%.2e\n",r);
   }

   release_wvd();

   if ((*status)>=0)
      (*status)/=nit;
   
   return 1.0/r;
}


static double lambda2(int nit)
{
   int Ns,nvh;
   int i,iw0,iw1;
   double r;
   complex_dble **wvd;

   Ns=Nss;
   nvh=nvhs;
   
   wvd=reserve_wvd(2);
   iw0=0;
   iw1=1;
   r=0.0;
   random_vd(nvh,wvd[iw0],1.0);
   Qvd(wvd[iw0],wvd[iw1]);
   vnormalize_dble(nvh,1,wvd[iw1]);

   for (i=0;i<nit;i++)
   {
      Zhat_dble(wvd[iw1],wvd[iw0]);
      Lvd(wvd[iw0],wvd[iw1]);
      r=vnormalize_dble(nvh,1,wvd[iw1]);

      message("r=%.2e\n",r);
   }

   release_wvd();
   
   return r;
}


static double lambda3(int nit,int nmx,double res,int *status)
{
   int Ns,nvh,stat;
   int i,iw0,iw1;
   double r;
   complex_dble **wvd;

   Ns=Nss;
   nvh=nvhs;

   wvd=reserve_wvd(2);
   iw0=0;
   iw1=1;
   (*status)=0;
   r=0.0;
   random_vd(nvh,wvd[iw0],1.0);
   Qvd(wvd[iw0],wvd[iw1]);
   vnormalize_dble(nvh,1,wvd[iw1]);

   for (i=0;i<nit;i++)
   {
      Zhatinv_dble(nmx,res,wvd[iw1],wvd[iw0],&stat);
      Qvd(wvd[iw0],wvd[iw1]);
      r=vnormalize_dble(nvh,1,wvd[iw1]);
      
      if ((*status)>=0)
      {
         if (stat>=0)
            (*status)+=stat;
         else
            (*status)=stat;
      }      

      message("r=%.2e\n",r);
   }

   release_wvd();

   if ((*status)>=0)
      (*status)/=nit;
   
   return 1.0/r;
}


int main(int argc,char *argv[])
{
   int my_rank,id;
   int bc,nc_first,nc_last;
   int bs_sap[4],nb,isw,nv;
   int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
   int vnkv,vnmx,vdnmx;   
   int nsize,icnfg,n,count;
   int status,statc,ifail,nfail,smin,smax,savg;
   double beta,kappa,csw;
   double rkappa,rvres,vres,vdres;   
   double r0,r1,r2,r3,wt1,wt2,tmin,tmax,tavg,rho,rmax;
   double c0min,c0max,c0avg,c1min,c1max,c1avg;
   complex_dble **wvd,z;
   char cnfg_dir[NAME_SIZE],cnfg_file[NAME_SIZE];
   char nbase[NAME_SIZE];   
   dfl_parms_t dfl;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check10.log","w",stdout);
      fin=freopen("check10.in","r",stdin);
      
      printf("\n");
      printf("Solution of the little Dirac equation\n");
      printf("-------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("id","%d\n",&id);
      read_line("cnfg_dir","%s\n",cnfg_dir);
      read_line("beta","%lf\n",&beta);
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);      
      read_line("bc","%d\n",&bc);      
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);      

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      
      read_line("bs_dfl","%d %d %d %d\n",
                &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
      read_line("Ns","%d\n",&Ns);
      read_line("rkappa","%lf\n",&rkappa);
      read_line("rinv","%d\n",&rinv);
      read_line("rnmr","%d\n",&rnmr);
      read_line("rncy","%d\n",&rncy);
      read_line("rvnmx","%d\n",&rvnmx);      
      read_line("rvres","%lf\n",&rvres);
      read_line("vnkv","%d\n",&vnkv);
      read_line("vnmx","%d\n",&vnmx);
      read_line("vres","%lf\n",&vres);
      read_line("vdnmx","%d\n",&vdnmx);
      read_line("vdres","%lf",&vdres);
 
      fclose(fin);

      printf("beta = %.6f\n",beta);
      printf("kappa = %.6f\n",kappa);
      printf("csw = %.6f\n",csw);
      printf("bc = %+d\n",bc);

      printf("bs_sap = %d %d %d %d\n",
             bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      
      printf("bs_dfl = %d %d %d %d\n",
             bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
      printf("Ns = %d\n",Ns);
      printf("rkappa = %.6f\n",rkappa);       
      printf("rinv = %d\n",rinv);
      printf("rnmr = %d\n",rnmr);       
      printf("rncy = %d\n",rncy);
      printf("rvnmx = %d\n",rvnmx);
      printf("rvres = %.1e\n",rvres);       
      printf("vnkv = %d\n",vnkv);
      printf("vnmx = %d\n",vnmx);
      printf("vres = %.1e\n",vres);
      printf("vdnmx = %d\n",vdnmx);
      printf("vdres = %.1e\n\n",vdres);
     
      printf("Configurations no %d -> %d\n\n",nc_first,nc_last);
      fflush(flog);      
   }

   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD);   

   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   
   MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   start_ranlux(0,123456);   
   geometry();

   alloc_u();
   alloc_ud();
   alloc_sw();
   alloc_swd();
   alloc_s(Ns+2);
   alloc_arrays(Ns);

   set_gcr_parms(bs_sap,1,1,1);
   dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                     rkappa,rvres,vres,vdres,0,5);
   blk_list(DFL_BLOCKS,&nb,&isw);
   nbs=nb;
   Nss=Ns;
   nvhs=(nb/2)*Ns;
   nv=nb*Ns;
   
   check_dir_root(cnfg_dir);   

   nsize=name_size("%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
                   NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   error_root(nsize>=NAME_SIZE,1,"main [check10.c]","Base name is too long");
   sprintf(nbase,"%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   nsize=name_size("%s/%sn%d",cnfg_dir,nbase,nc_last);
   error_root(nsize>=NAME_SIZE,1,"main [check10.c]",
              "cnfg_dir name is too long");

   nfail=0;
   smin=0;
   smax=0;
   savg=0;
   tmin=0.0;
   tmax=0.0;
   tavg=0.0;
   rmax=0.0;
   c0min=0.0;
   c0max=0.0;
   c0avg=0.0;
   c1min=0.0;
   c1max=0.0;
   c1avg=0.0;
   
   for (icnfg=nc_first;icnfg<=nc_last;icnfg++)
   {
      if (my_rank==0)
      {
         printf("Configuration no %d:\n\n",icnfg);
         fflush(flog);
      }

      set_lat_parms(beta,kappa,csw);
      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);
      if (bc==-1)
         flipbcd();
      dfl_modes(0,&status,&ifail);      

      error_root((status<0)||(ifail!=0),1,"main [check10.c]",
                 "dfl_modes failed (status=%d, ifail=%d)",status,ifail);

      dfl_little_modes(vs,zvs,vds,zvds,avd);
      
      wvd=reserve_wvd(3);
      random_vd(nv,wvd[0],1.0);
      assign_vd2vd(nv,wvd[0],wvd[2]);
      dfl_solve_little(vdnmx,vdres,wvd[0],&status);
      Znohat_dble(wvd[0],wvd[1]);
      z.re=-1.0;
      z.im=0.0;
      mulc_vadd_dble(nv,wvd[1],wvd[2],z);
      rho=vnorm_square_dble(nv,1,wvd[1])/vnorm_square_dble(nv,1,wvd[2]);
      rho=sqrt(rho);

      message("Solution for a random source:\n");
      message("Status = %d, relative residue = %.1e "
              "(should be less than %.1e)\n\n",status,rho,vdres);

      n=(int)(1.0e9/(64.0*(double)(Ns*nv*status)));
      if (n<1)
         n=1;
      
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();
      for (count=0;count<n;count++)
      {
         assign_vd2vd(nv,wvd[2],wvd[0]);
         dfl_solve_little(vdnmx,vdres,wvd[0],&status);
      }
      MPI_Barrier(MPI_COMM_WORLD);
      wt2=(MPI_Wtime()-wt1)/(double)(n);

      message("Timing of dfl_solve_little:\n");
      message("Time per solution          = %.3f sec\n",
              wt2);
      message("Time per solution & point  = %.3f msec\n",
              1.0e3*wt2/(double)(VOLUME));
      message("Time per iteration         = %.3f msec\n",
              1.0e3*wt2/(double)(status));
      message("Time per iteration & point = %.3f usec\n\n",
              1.0e6*wt2/(double)(VOLUME*status));
      
      release_wvd();
      
      if (my_rank==0)
         fflush(flog);

      message("Computation of the condition number of Zhat:\n");
      
      r0=lambda0(20);
      message("Maximal eigenvalue = %.1e\n",r0);
      r1=lambda1(20,vnmx,vres,&statc);
      message("Minimal eigenvalue = %.1e (status = %d)\n",r1,statc);
      r0/=r1;
      message("Condition number   = %.1e\n\n",r0);      

      message("Computation of the condition number of PL*Zhat*(1-P):\n");
      
      r2=lambda2(20);
      message("Maximal eigenvalue = %.1e\n",r2);
      r3=lambda3(20,vnmx,vres,&statc);
      message("Minimal eigenvalue = %.1e (status = %d)\n",r3,statc);
      r2/=r3;
      message("Condition number   = %.1e\n\n",r2);

      if (status<0)
         nfail+=1;
      else
      {
         if (icnfg==(nc_first+nfail))
         {
            smin=status;
            smax=status;
            savg=status;

            tmin=wt2;
            tmax=wt2;
            tavg=wt2;

            rmax=rho;

            c0min=r0;
            c0max=r0;
            c0avg=r0;
            
            c1min=r2;
            c1max=r2;
            c1avg=r2;            
         }
         else
         {
            if (status<smin)
               smin=status;
            if (status>smax)
               smax=status;
            savg+=status;

            if (wt2<tmin)
               tmin=wt2;
            if (wt2>tmax)
               tmax=wt2;
            tavg+=wt2;

            if (rho>rmax)
               rmax=rho;
            
            if (r0<c0min)
               c0min=r0;
            if (r0>c0max)
               c0max=r0;
            c0avg+=r0;

            if (r2<c1min)
               c1min=r2;
            if (r2>c1max)
               c1max=r2;
            c1avg+=r2;
         }
      }
      
      error_chk();
   }
      
   if (my_rank==0)
   {
      printf("Test summary\n");
      printf("------------\n\n");

      n=(nc_last-nc_first+1-nfail);

      if (nfail==0)
         printf("There were no failures\n");
      else
         printf("There were %d failures\n",nfail);

      printf("Maximal relative residue = %.1e (should be less than %.1e)\n\n",
             rmax,vdres);

      printf("dfl_solve_little:\n");
      printf("Iteration number  (min,max,avg) = %3d, %3d, %3d\n",
             smin,smax,savg/n);
      printf("Time per solution (min,max,avg) = %.3f, %.3f, %.3f sec\n",
             tmin,tmax,tavg/(double)(n));
      printf("Average per solution & point    = %.3f msec\n",
             1.0e3*tavg/(double)(n*VOLUME));
      printf("Average per iteration           = %.3f msec\n",
             1.0e3*tavg/(double)(savg));
      printf("Average per iteration & point   = %.3f usec\n\n",
             1.0e6*tavg/(double)(VOLUME*savg));      

      printf("Condition number of Zhat (min,max,avg) = %3d, %3d, %3d\n",
             (int)(c0min),(int)(c0max),(int)(c0avg/(double)(n)));
      printf("Condition number of PL*Zhat*(1-P)      = %3d, %3d, %3d\n\n",
             (int)(c1min),(int)(c1max),(int)(c1avg/(double)(n)));
      fclose(flog);
   }
   
   MPI_Finalize();   
   exit(0);
}
