
/*******************************************************************************
*
* File check6.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Consistency of Zhat_dble(), Zeeinv_dble(), ... , Znohat()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "flags.h"
#include "block.h"
#include "linalg.h"
#include "dfl.h"
#include "global.h"


static void random_basis(int Ns)
{
   int i;
   spinor **ws;

   ws=reserve_ws(Ns);

   for (i=0;i<Ns;i++)
      random_s(VOLUME,ws[i],1.0f);
   
   dfl_subspace(ws);

   release_ws();
}


static void check_vd(int Ns,int nvh,complex_dble **vd)
{
   int i,j;
   double d,dev;
   complex_dble z;

   dev=0.0;

   for (i=0;i<Ns;i++)
   {
      for (j=0;j<i;j++)
      {
         z=vprod_dble(nvh,1,vd[i],vd[j]);
         d=sqrt(z.re*z.re+z.im*z.im);

         if (i==j)
            d=fabs(1.0-d);

         if (d>dev)
            dev=d;
      }
   }   

   message("Orthonormality of the vd fields:  %.1e\n",dev); 
}


static void check_zvd(int Ns,int nvh,complex_dble **vd,complex_dble **zvd)
{
   int i;
   double d,dev;
   complex_dble z,**wvd;

   dev=0.0;
   z.re=-1.0;
   z.im=0.0;

   wvd=reserve_wvd(1);

   for (i=0;i<Ns;i++)
   {
      Zhat_dble(vd[i],wvd[0]);
      mulc_vadd_dble(nvh,wvd[0],zvd[i],z);

      d=vnorm_square_dble(nvh,1,wvd[0]);

      if (d>dev)
         dev=d;
   }   

   release_wvd();
   
   message("Check of the zvd fields:          %.1e\n",sqrt(dev)); 
}


static void check_a(int Ns,int nvh,complex_dble **vd,complex_dble **zvd,
                    complex_dble *amat,complex_dble *bmat,complex_dble *cmat)
{
   int i,j;
   double d,dev;
   
   for (i=0;i<Ns;i++)
   {
      for (j=0;j<Ns;j++)
         bmat[i*Ns+j]=vprod_dble(nvh,1,vd[i],zvd[j]);
   }

   cmat_mul_dble(Ns,amat,bmat,cmat);
   dev=0.0;

   for (i=0;i<Ns;i++)
   {
      for (j=0;j<Ns;j++)
      {
         if (i==j)
            d=fabs(1.0-cmat[i*Ns+j].re)+fabs(cmat[i*Ns+j].im);
         else
            d=fabs(cmat[i*Ns+j].re)+fabs(cmat[i*Ns+j].im);

         if (d>dev)
            dev=d;
      }
   }

   if (NPROC>1)
   {
      MPI_Reduce(&dev,&d,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(&d,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
      dev=d;
   }

   message("Little little  matrix:            %.1e\n",dev);
}


static void check_vzv(int Ns,int nvh,complex **v,complex **zv,
                      complex_dble **vd,complex_dble **zvd)
{
   int i;
   double d,dev;
   complex_dble z,**wvd;

   dev=0.0;
   z.re=-1.0;
   z.im=0.0;

   wvd=reserve_wvd(1);

   for (i=0;i<Ns;i++)
   {
      assign_v2vd(nvh,v[i],wvd[0]);
      mulc_vadd_dble(nvh,wvd[0],vd[i],z);
      d=vnorm_square_dble(nvh,1,wvd[0]);

      if (d>dev)
         dev=d;

      assign_v2vd(nvh,zv[i],wvd[0]);
      mulc_vadd_dble(nvh,wvd[0],zvd[i],z);
      d=vnorm_square_dble(nvh,1,wvd[0]);

      if (d>dev)
         dev=d;
   }

   release_wvd();
   
   message("Check of v and zv:                %.1e\n",sqrt(dev));   
}


int main(int argc,char *argv[])
{
   int my_rank,iop;
   int bs[4],Ns,nb,nv,nvh;   
   double d;
   complex **wv,**v,**zv,*smat,z;
   complex_dble **wvd,**vd,**zvd,zd;
   complex_dble *amat,*bmat,*cmat;
   void (*op)(complex *v,complex *w);   
   void (*op_dble)(complex_dble *v,complex_dble *w);
   char *pr,*prd;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check6.log","w",stdout);
      fin=freopen("check6.in","r",stdin);
      
      printf("\n");
      printf("Consistency of Zhat_dble(), Zeeinv_dble(), ..., Zhat()\n");
      printf("------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("Ns","%d",&Ns);
      fclose(fin);

      if (Ns<2)
         Ns=2;
      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("Ns = %d\n\n",Ns);
      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   

   start_ranlux(0,123456);   
   geometry();

   alloc_ud();
   alloc_s(Ns);

   set_lat_parms(0.0,0.1,1.0);
   set_dfl_parms(bs,Ns,4,1,1,1,1,1,1,0.12,1.0e-3,1.0e-6,1.0e-12,4,5);
   
   random_ud();
   random_basis(Ns);
   
   nb=VOLUME/(bs[0]*bs[1]*bs[2]*bs[3]);
   nv=nb*Ns;
   nvh=nv/2;
   wv=reserve_wv(4);
   wvd=reserve_wvd(4);
   zd.re=-1.0;
   zd.im=0.0;
   z.re=-1.0f;
   z.im=0.0f;

   for (iop=0;iop<6;iop++)
   {
      if (iop==0)
      {
         op=Zhat;
         pr= "Zhat()  ";
         op_dble=Zhat_dble;
         prd="Zhat_dble()  ";
      }
      else if (iop==1)
      {
         op=Zeeinv;
         pr= "Zeeinv()";         
         op_dble=Zeeinv_dble;
         prd="Zeeinv_dble()";
      }
      else if (iop==2)
      {
         op=Zooinv;
         pr= "Zooinv()";         
         op_dble=Zooinv_dble;
         prd="Zooinv_dble()";
      }
      else if (iop==3)
      {
         op=Zoe;
         pr= "Zoe()   ";         
         op_dble=Zoe_dble;
         prd="Zoe_dble()   ";
      }      
      else if (iop==4)
      {
         op=Zeo;
         pr= "Zeo()   ";         
         op_dble=Zeo_dble;
         prd="Zeo_dble()   ";
      }      
      else
      {
         op=Znohat;
         pr= "Znohat()";         
         op_dble=Znohat_dble;
         prd="Znohat_dble()";
      }
      
      random_vd(nv,wvd[0],1.0);
      random_vd(nv,wvd[1],1.0);      
      assign_vd2vd(nv,wvd[0],wvd[2]);
      assign_vd2vd(nv,wvd[1],wvd[3]);

      assign_vd2v(nv,wvd[0],wv[0]);
      assign_vd2v(nv,wvd[1],wv[1]);
      assign_v2v(nv,wv[0],wv[2]);
      assign_v2v(nv,wv[1],wv[3]);

      op_dble(wvd[0],wvd[1]);
      op(wv[0],wv[1]);

      mulc_vadd_dble(nv,wvd[2],wvd[0],zd);
      d=vnorm_square_dble(nv,0,wvd[2]);
      error(d!=0.0,1,"main [check6.c]",
            "%s modifies the input field",prd);

      mulc_vadd(nv,wv[2],wv[0],z);
      d=(double)(vnorm_square(nv,0,wv[2]));
      error(d!=0.0,1,"main [check6.c]",
            "%s modifies the input field",pr);      

      if ((iop<2)||(iop==4))
      {
         mulc_vadd_dble(nvh,wvd[3]+nvh,wvd[1]+nvh,zd);
         d=vnorm_square_dble(nvh,0,wvd[3]+nvh);
         error(d!=0.0,1,"main [check6.c]",
               "%s modifies the odd components of the output field",prd);

         mulc_vadd(nvh,wv[3]+nvh,wv[1]+nvh,z);
         d=(double)(vnorm_square(nvh,0,wv[3]+nvh));
         error(d!=0.0,1,"main [check6.c]",
               "%s modifies the odd components of the output field",pr);

         assign_vd2v(nvh,wvd[1],wv[0]);
         mulc_vadd(nvh,wv[0],wv[1],z);
         d=(double)(vnorm_square(nvh,1,wv[0])/
                    vnorm_square(nvh,1,wv[1]));
         message("Deviation of %s from %s: %.1e\n",pr,prd,sqrt(d));
      }

      if ((iop==2)||(iop==3))
      {
         mulc_vadd_dble(nvh,wvd[3],wvd[1],zd);
         d=vnorm_square_dble(nvh,0,wvd[3]);
         error(d!=0.0,1,"main [check6.c]",
               "%s modifies the even components of the output field",prd);

         mulc_vadd(nvh,wv[3],wv[1],z);
         d=(double)(vnorm_square(nvh,0,wv[3]));
         error(d!=0.0,1,"main [check6.c]",
               "%s modifies the even components of the output field",pr);

         assign_vd2v(nvh,wvd[1]+nvh,wv[0]+nvh);
         mulc_vadd(nvh,wv[0]+nvh,wv[1]+nvh,z);
         d=(double)(vnorm_square(nvh,1,wv[0]+nvh)/
                    vnorm_square(nvh,1,wv[1]+nvh));
         message("Deviation of %s from %s: %.1e\n",pr,prd,sqrt(d));         
      }

      if (iop==5)
      {
         assign_vd2v(nv,wvd[1],wv[0]);
         mulc_vadd(nv,wv[0],wv[1],z);
         d=(double)(vnorm_square(nv,1,wv[0])/
                    vnorm_square(nv,1,wv[1]));
         message("Deviation of %s from %s: %.1e\n",pr,prd,sqrt(d));
      }
   }

   flipbcd();
   random_basis(Ns);
   
   random_vd(nvh,wvd[0],1.0);
   set_vd2zero(nvh,wvd[0]+nvh);
   Znohat_dble(wvd[0],wvd[1]);

   Zeeinv_dble(wvd[1],wvd[2]);
   mulc_vadd_dble(nvh,wvd[2],wvd[0],zd);
   d=vnorm_square_dble(nvh,1,wvd[2])/vnorm_square_dble(nvh,1,wvd[0]);   
   message("\nComparison of Zeeinv_dble() and Znohat_dble(): %.1e\n",sqrt(d));
   
   Zoe_dble(wvd[0],wvd[2]);
   mulc_vadd_dble(nvh,wvd[2]+nvh,wvd[1]+nvh,zd);
   d=vnorm_square_dble(nvh,1,wvd[2]+nvh)/vnorm_square_dble(nvh,1,wvd[1]+nvh);   
   message("Comparison of Zoe_dble()    and Znohat_dble(): %.1e\n",sqrt(d));

   random_vd(nvh,wvd[0]+nvh,1.0);
   set_vd2zero(nvh,wvd[0]);
   Znohat_dble(wvd[0],wvd[1]);

   Zooinv_dble(wvd[1],wvd[2]);
   mulc_vadd_dble(nvh,wvd[2]+nvh,wvd[0]+nvh,zd);
   d=vnorm_square_dble(nvh,1,wvd[2]+nvh)/vnorm_square_dble(nvh,1,wvd[0]+nvh);   
   message("Comparison of Zooinv_dble() and Znohat_dble(): %.1e\n",sqrt(d));

   random_vd(nvh,wvd[2],1.0);
   assign_vd2vd(nvh,wvd[2],wvd[3]);
   Zeo_dble(wvd[0],wvd[2]);
   mulc_vadd_dble(nvh,wvd[3],wvd[2],zd);   
   mulc_vadd_dble(nvh,wvd[3],wvd[1],zd);
   d=vnorm_square_dble(nvh,1,wvd[3])/vnorm_square_dble(nvh,1,wvd[1]);   
   message("Comparison of Zeo_dble()    and Znohat_dble(): %.1e\n",sqrt(d));   

   random_vd(nv,wvd[0],1.0);
   Zhat_dble(wvd[0],wvd[1]);
   Zoe_dble(wvd[0],wvd[2]);
   Zooinv_dble(wvd[2],wvd[3]);
   set_vd2zero(nvh,wvd[0]+nvh);
   Znohat_dble(wvd[0],wvd[2]);
   Zeo_dble(wvd[3],wvd[2]);
   Zeeinv_dble(wvd[2],wvd[3]);
   
   mulc_vadd_dble(nvh,wvd[3],wvd[1],zd);
   d=vnorm_square_dble(nvh,1,wvd[3])/vnorm_square_dble(nvh,1,wvd[1]);
   message("Comparison of Zeeinv_dble(), Zooinv_dble(), \n");
   message("       Zoe_dble(), Zeo_dble() and Zhat_dble(): %.1e\n\n",sqrt(d));

   v=amalloc(2*Ns*sizeof(*v),3);
   vd=amalloc(2*Ns*sizeof(*vd),3);
   amat=amalloc(3*Ns*Ns*sizeof(*amat),ALIGN);
   smat=amalloc(Ns*Ns*sizeof(*smat),ALIGN);
   bmat=amat+Ns*Ns;
   cmat=bmat+Ns*Ns;
   error((v==NULL)||(vd==NULL)||(smat==NULL)||(amat==NULL),1,
         "main [check6.c]","Unable to allocate auxiliary arrays");

   zv=v+Ns;
   zvd=vd+Ns;

   message("Check of the little dfl modes:\n");
   
   dfl_little_modes(v,zv,vd,zvd,amat);
   check_vd(Ns,nvh,vd);
   check_zvd(Ns,nvh,vd,zvd);
   check_a(Ns,nvh,vd,zvd,amat,bmat,cmat);
   check_vzv(Ns,nvh,v,zv,vd,zvd);

   error_chk();
   
   if (my_rank==0)
      fclose(flog);
   
   MPI_Finalize();   
   exit(0);
}
