
/*******************************************************************************
*
* File time2.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of s2v and sub_v2s
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "sw_term.h"
#include "dfl.h"
#include "global.h"


static void random_basis(int Ns)
{
   int i;
   spinor **ws;

   ws=reserve_ws(Ns);

   for (i=0;i<Ns;i++)
      random_s(VOLUME,ws[i],1.0f);
   
   dfl_subspace(ws);

   release_ws();
}


int main(int argc,char *argv[])
{
   int my_rank,nflds;
   int bs[4],Ns,nb,isw,nv,iws0;
   int n,i,count;
   double wt1,wt2,wdt;
   complex **wv;
   spinor **ws;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time2.log","w",stdout);
      fin=freopen("time2.in","r",stdin);
      
      printf("\n");
      printf("Timing of s2v and sub_v2s\n");
      printf("-------------------------\n\n");   
      
      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("Ns","%d",&Ns);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("Ns = %d\n\n",Ns);
      fflush(flog);      
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   

   start_ranlux(0,12345);
   geometry();

   alloc_s(Ns);

   set_lat_parms(0.0,0.12,1.0);
   set_dfl_parms(bs,Ns,4,1,1,1,1,1,1,0.12,1.0e-3,1.0e-6,1.0e-12,0,0);
   
   blk_list(DFL_BLOCKS,&nb,&isw);
   nv=nb*Ns;

   if (my_rank==0)
   {
      printf("Number of points = %d\n",VOLUME);
      printf("Number of blocks = %d\n",nb);
      printf("Number of points/block = %d\n",bs[0]*bs[1]*bs[2]*bs[3]);
      printf("Vector field size = %.2f KB\n",
             (double)(sizeof(complex)*nb*Ns)*1.0e-3);
      printf("Spinor field size = %.2f MB\n\n",
             (double)(sizeof(spinor)*VOLUME)*1.0e-6);
   }

   nflds=(int)(4.0e6/(double)(sizeof(spinor)*VOLUME));
   if ((nflds%2)!=0)
      nflds+=1;
   if (nflds==0)
      nflds=2;

   alloc_s(Ns+nflds);   
   set_dfl_parms(bs,Ns,4,1,1,1,1,1,1,0.12,1.0e-3,1.0e-6,1.0e-12,nflds,0);   
   
   random_basis(Ns);
   ws=reserve_ws(nflds);
   wv=reserve_wv(nflds);
   iws0=(ws[0]-ps[0][0])/NSPIN;
   
   for (i=0;i<nflds;i++)
   {
      random_s(VOLUME,ws[i],1.0f);
      random_v(nv,wv[i],1.0f);
   }

   n=1.0e3/(double)(Ns);
   if (n<1)
      n=1;
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime(); 
   for (count=0;count<n;count++)
   {
      for (i=0;i<nflds;i+=2)
         dfl_s2v(iws0+i,wv[i]);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   
   wdt=2.0e6*(wt2-wt1)/(double)(n*nflds);

   if (my_rank==0)
   {   
      printf("Timing of dfl_s2v():\n");
      printf("Time per point: %4.3f micro sec\n\n",wdt/(double)(VOLUME));
   }

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime(); 
   for (count=0;count<n;count++)
   {
      for (i=1;i<nflds;i+=2)
         dfl_sub_v2s(wv[i],iws0+i);
   }
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   
   wdt=2.0e6*(wt2-wt1)/(double)(n*nflds);

   if (my_rank==0)
   {   
      printf("Timing of dfl_sub_v2s():\n");
      printf("Time per point: %4.3f micro sec\n\n",wdt/(double)(VOLUME));
      fclose(flog);
   }   
   
   MPI_Finalize();   
   exit(0);
}
