
/*******************************************************************************
*
* File check6.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of hmc_acf() and hmc_frf()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "linalg.h"
#include "dirac.h"
#include "sw_term.h"
#include "hmc.h"
#include "global.h"

static const su3_dble ud0={{0.0}};


static void init_ud(void)
{
   su3_alg_dble X;
   su3_dble *u,*um;

   u=pud[VOLUME/2][0];
   um=u+4*VOLUME;

   for (;u<um;u++)
   {
      *u=ud0;
      (*u).c11.re=1.0;
      (*u).c22.re=1.0;
      (*u).c33.re=1.0;      
      random_alg(1,&X);
      ExpX(0.1,&X,u);      
   }

   set_flags(NEW_UD);
   renormalize_ud();
}


static void rot_ud(int n,double t)
{
   int nb,isw,*lk,*lkm;
   su3_dble *ub;
   su3_alg_dble *mom;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;
   hflds=hmcflds()+n;
   
   ub=(*b).ud[0];
   mom=(*hflds).mom;
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
   {
      ExpX(t,mom,ub+(*lk));
      mom+=1;
   }

   set_blk_flags(HMC_BLOCKS,NEW_ACTIVE_UD,n);
}


static double dSdt(int n)
{
   double sm;
   su3_alg_dble *fr,*frm,*mom;
   hmc_t *hflds;

   hflds=hmcflds()+n;

   fr=(*hflds).frc;
   frm=fr+(*hflds).nlks;
   mom=(*hflds).mom;   
   sm=0.0;
   
   for (;fr<frm;fr++)
   {
      sm+=12.0*((*mom).c1*(*fr).c1+(*mom).c2*(*fr).c2);
      sm-= 6.0*((*mom).c1*(*fr).c2+(*mom).c2*(*fr).c1);
      sm+= 4.0*((*mom).c3*(*fr).c3+(*mom).c4*(*fr).c4+(*mom).c5*(*fr).c5+
                (*mom).c6*(*fr).c6+(*mom).c7*(*fr).c7+(*mom).c8*(*fr).c8);
      mom+=1;
   }

   return sm;
}


static void mul_swde_inv(block_t *b,int k)
{
   int ifail;
   weyl_dble *s,*sm;
   pauli_dble *m,*minv;

   minv=amalloc(sizeof(pauli_dble),ALIGN);
   error(minv==NULL,1,"mul_swde_inv [check6.c]",
         "Unable to allocate workspace");

   ifail=0;
   s=(weyl_dble*)((*b).sd[k]);
   sm=s+(*b).vol;
   m=(*b).swd;

   for (;s<sm;s++)
   {
      ifail|=inv_pauli_dble(m,minv);
      mul_pauli_dble(minv,s,s);
      m+=1;
   }

   error(ifail!=0,1,"mul_swde_inv [check6.c]",
         "Unsafe inversion of the SW-term"); 
   afree(minv);
}


static double sw_act(int n)
{
   int nb,isw;
   double act,c,p;
   pauli_dble *m,*mm,*mn;
   block_t *b;
   lat_parms_t lat;

   if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)==0)||
       (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n))||
       (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)))       
      blk_sw_term(n);

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;

   lat=lat_parms();
   c=1.0/(4.0+lat.m0);
   c=c*c;
   c=c*c*c;

   act=0.0;   
   m=(*b).swd;
   mm=m+2*(*b).vol;

   for (;m<mm;)
   {
      mn=m+16;
      p=1.0;

      for (;m<mn;m++)
         p*=(c*det_pauli_dble(m));

      act-=log(fabs(p));
   }

   return 2.0*act;
}


int main(int argc,char *argv[])
{
   int my_rank,nb,isw,n,i;
   int bc,bs_hmc[4],N0,N1,N2;
   int nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf;
   int status,ifail,flag;
   int stat_acf_max,stat_acf_min;
   int stat_frf_max,stat_frf_min;
   double beta,kappa,csw;
   double tau,res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf;
   double eps,dsdt,act0,act1,dact,dact_max;
   double dev_frc,dev_frc_all,dev_frc_max;
   double sig_loss,sig_loss_all,sig_loss_max;
   block_t *b;
   hmc_t *hflds;
   lat_parms_t lat;
   hmc_parms_t hmc;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check6.log","w",stdout);
      fin=freopen("check6.in","r",stdin);
      
      printf("\n");
      printf("Check of hmc_acf() and hmc_frf()\n");
      printf("--------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);
      read_line("bc","%d\n",&bc);
      read_line("bs_hmc","%d %d %d %d\n",
                &bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      read_line("nmx_acf","%d\n",&nmx_acf);
      read_line("res_acf","%lf\n",&res_acf);
      read_line("nmx_frf","%d\n",&nmx_frf);
      read_line("res_frf","%lf",&res_frf);      
      fclose(fin);

      printf("Random gauge field\n");      
   }

   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);   
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_acf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_acf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx_frf,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res_frf,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   start_ranlux(0,1234);
   geometry();
   alloc_ud();

   beta=0.0;
   N0=1;
   N1=1;
   N2=1;
   nmx_gcr_acf=1;
   nmx_rop=1;
   nmx_gcr_frf=1;
   tau=0.0;
   res_gcr_acf=1.0;
   res_rop=1.0;
   res_gcr_frf=1.0;

   lat=set_lat_parms(beta,kappa,csw);
   hmc=set_hmc_parms(bs_hmc,bc,N0,N1,N2,0,0,0,3,
                     nmx_acf,nmx_gcr_acf,nmx_rop,nmx_frf,nmx_gcr_frf,
                     tau,
                     res_acf,res_gcr_acf,res_rop,res_frf,res_gcr_frf);

   if (my_rank==0)
   {
      printf("kappa = %.6f\n",lat.kappa);      
      printf("csw = %.6f\n",lat.csw);
      printf("bc = %d\n",hmc.bc);
      printf("cno_blk = %d\n",hmc.cno_blk);
      printf("bs_hmc = %d %d %d %d\n",bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("nmx_acf = %d\n",hmc.nmx_acf);            
      printf("res_acf = %.2e\n",hmc.res_acf);      
      printf("nmx_frf = %d\n",hmc.nmx_frf);            
      printf("res_frf = %.2e\n\n",hmc.res_frf);      
   }

   init_ud();
   if (bc==-1)
      flipbcd();   
   assign_ud2udbgr(HMC_BLOCKS);
   swb_term();

   ifail=0;
   stat_acf_max=0;
   stat_acf_min=0;
   dact_max=0.0;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   for (n=0;n<nb;n++)
   {
      random_sd((*b).volh,(*b).sd[0],1.0);
      act0=norm_square_dble((*b).volh,0,(*b).sd[0]);

      ifail|=invert_blk_swd(HMC_BLOCKS,n,ODD_PTS);

      Qhat_blk_dble(b,0,1);
      mul_swde_inv(b,1);
      assign_sd2sd((*b).volh,(*b).sd[1],(*hflds).phi);

      act1=hmc_acf(n,1,0,&status,&flag);
      ifail|=flag;

      dact=fabs(1.0-act1/act0);
      if (status>stat_acf_max)
         stat_acf_max=status;
      if ((stat_acf_min==0)||(status<stat_acf_min))
         stat_acf_min=status;
      if (dact>dact_max)
         dact_max=dact;
   
      b+=1;      
      hflds+=1;
   }

   error_chk();
   MPI_Reduce(&stat_acf_max,&status,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   stat_acf_max=status;
   MPI_Reduce(&stat_acf_min,&status,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
   stat_acf_min=status;
   MPI_Reduce(&dact_max,&dact,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   dact_max=dact;
   MPI_Reduce(&ifail,&flag,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   ifail=flag;
   
   if (my_rank==0)
   {
      printf("Calculation of the action\n");
      printf("Pseudo-fermions:\n");
      printf("relative deviation = %.2e\n",dact_max);
      printf("status             = %d .. %d, ifail = %d\n",
             stat_acf_min,stat_acf_max,ifail);
   }

   ifail=0;
   stat_acf_max=0;
   stat_acf_min=0;
   dact_max=0.0;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   for (n=0;n<nb;n++)
   {
      act0=sw_act(n);
      act1=hmc_acf(n,0,1,&status,&flag);
      ifail|=flag;

      dact=fabs(1.0-act1/act0);
      if (status>stat_acf_max)
         stat_acf_max=status;
      if ((stat_acf_min==0)||(status<stat_acf_min))
         stat_acf_min=status;
      if (dact>dact_max)
         dact_max=dact;
   
      b+=1;      
      hflds+=1;
   }

   error_chk();
   MPI_Reduce(&stat_acf_max,&status,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   stat_acf_max=status;
   MPI_Reduce(&stat_acf_min,&status,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
   stat_acf_min=status;
   MPI_Reduce(&dact_max,&dact,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   dact_max=dact;
   MPI_Reduce(&ifail,&flag,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   ifail=flag;
   
   if (my_rank==0)
   {
      printf("Determinants:\n");
      printf("relative deviation = %.2e\n",dact_max);
      printf("status             = %d .. %d, ifail = %d\n",
             stat_acf_min,stat_acf_max,ifail);
   }

   ifail=0;
   stat_acf_max=0;
   stat_acf_min=0;
   dact_max=0.0;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   for (n=0;n<nb;n++)
   {
      random_sd((*b).volh,(*b).sd[0],1.0);
      act0=norm_square_dble((*b).volh,0,(*b).sd[0])+sw_act(n);

      ifail|=invert_blk_swd(HMC_BLOCKS,n,ODD_PTS);

      Qhat_blk_dble(b,0,1);
      mul_swde_inv(b,1);
      assign_sd2sd((*b).volh,(*b).sd[1],(*hflds).phi);

      act1=hmc_acf(n,1,1,&status,&flag);
      ifail|=flag;

      dact=fabs(1.0-act1/act0);
      if (status>stat_acf_max)
         stat_acf_max=status;
      if ((stat_acf_min==0)||(status<stat_acf_min))
         stat_acf_min=status;
      if (dact>dact_max)
         dact_max=dact;
   
      b+=1;      
      hflds+=1;
   }

   error_chk();
   MPI_Reduce(&stat_acf_max,&status,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   stat_acf_max=status;
   MPI_Reduce(&stat_acf_min,&status,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
   stat_acf_min=status;
   MPI_Reduce(&dact_max,&dact,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   dact_max=dact;
   MPI_Reduce(&ifail,&flag,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   ifail=flag;
   
   if (my_rank==0)
   {
      printf("Both:\n");
      printf("relative deviation = %.2e\n",dact_max);
      printf("status             = %d .. %d, ifail = %d\n\n",
             stat_acf_min,stat_acf_max,ifail);
   }

   for (i=0;i<6;i++)
   {
      ifail=0;
      stat_acf_max=0;
      stat_acf_min=0;
      stat_frf_max=0;
      stat_frf_min=0;      
      dev_frc_max=0.0;
      sig_loss_max=0.0;
      eps=3.0e-4;

      b=blk_list(HMC_BLOCKS,&nb,&isw);
      hflds=hmcflds();   

      for (n=0;n<nb;n++)
      {
         if ((i>0)&&(i<5))
            rot_ud(n,eps);
         if (i==0)
            random_alg((*hflds).nlks,(*hflds).mom);

         hmc_frf(n,&status,&flag);

         ifail|=flag;         
         if (status>stat_frf_max)
            stat_frf_max=status;
         if ((stat_frf_min==0)||(status<stat_frf_min))
            stat_frf_min=status;

         dsdt=dSdt(n);

         rot_ud(n,eps);
         act0=2.0*hmc_acf(n,1,1,&status,&flag)/3.0;
         rot_ud(n,-eps);

         ifail|=flag;
         if (status>stat_acf_max)
            stat_acf_max=status;
         if ((stat_acf_min==0)||(status<stat_acf_min))
            stat_acf_min=status;
      
         rot_ud(n,-eps);
         act1=2.0*hmc_acf(n,1,1,&status,&flag)/3.0;
         rot_ud(n,eps);

         ifail|=flag;
         if (status>stat_acf_max)
            stat_acf_max=status;
         if ((stat_acf_min==0)||(status<stat_acf_min))
            stat_acf_min=status;
      
         rot_ud(n,2.0*eps);
         act0-=hmc_acf(n,1,1,&status,&flag)/12.0;
         rot_ud(n,-2.0*eps);

         ifail|=flag;
         if (status>stat_acf_max)
            stat_acf_max=status;
         if ((stat_acf_min==0)||(status<stat_acf_min))
            stat_acf_min=status;
      
         rot_ud(n,-2.0*eps);
         act1-=hmc_acf(n,1,1,&status,&flag)/12.0;
         rot_ud(n,2.0*eps);

         ifail|=flag;
         if (status>stat_acf_max)
            stat_acf_max=status;
         if ((stat_acf_min==0)||(status<stat_acf_min))
            stat_acf_min=status;
      
         dact=(act0-act1)/eps;
         dev_frc=fabs(1.0-dsdt/dact);
         if (dev_frc>dev_frc_max)
            dev_frc_max=dev_frc;

         sig_loss=-log10(fabs(1.0-act0/act1));
         if (sig_loss>sig_loss_max)
            sig_loss_max=sig_loss;      

         b+=1;
         hflds+=1;
      }

      error_chk();
      MPI_Reduce(&stat_frf_max,&status,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
      stat_frf_max=status;
      MPI_Reduce(&stat_frf_min,&status,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
      stat_frf_min=status;      
      MPI_Reduce(&stat_acf_max,&status,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
      stat_acf_max=status;
      MPI_Reduce(&stat_acf_min,&status,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
      stat_acf_min=status;
      MPI_Reduce(&ifail,&flag,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
      ifail=flag;
      MPI_Reduce(&dev_frc_max,&dev_frc_all,1,
                 MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Reduce(&sig_loss_max,&sig_loss_all,1,
                 MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);

      if (my_rank==0)
      {
         if (i==0)
            printf("Calculation of the force:\n\n");

         if (i<3)
            printf("nold = %d ",i);
         else
            printf("nold = %d ",3);

         if (i==5)
            printf("(field unchanged)\n");
         else
            printf("\n");
         
         printf("Relative deviation of dS/dt = %.2e ",dev_frc_all);
         printf("[significance loss = %d digits]\n",(int)(sig_loss_all));
         printf("status (force) = %d .. %d, status (action) = %d .. %d, "
                "ifail = %d\n\n",stat_frf_min,stat_frf_max,
                stat_acf_min,stat_acf_max,ifail);
         fflush(flog);
      }
   }

   if (my_rank==0)
      fclose(flog);
   
   MPI_Finalize();    
   exit(0);
}
