
/*******************************************************************************
*
* File lpfrog.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Leap-frog integration of the molecular-dynamics (MD) equations
*
* The externally accessible functions are
*
*   void alloc_hmc_action(hmc_act_t *act)
*     Allocates the data arrays in the action structure pointed to by act
*     and initializes them to zero. The size of the arrays is determined
*     by the number nb of blocks in the HMC_BLOCKS grid and (*act).nb is
*     set to nb (see the notes)
*
*   void start_md(hmc_act_t act0,int *stat_rop,int *stat_dfl_mds,int *ifail)
*     Initialization of the momentum fields, the pseudo-fermion fields, the
*     gauge fields on the blocks of the HMC_BLOCKS grid and of the deflation
*     subspace (if deflation is switched on). The program also computes the
*     active parts of the HMC action and assigns the result to the elements
*     of the structure act0 (see the notes).
*      The parameters stat_* point to a single value (see the notes). The
*     SW-term is updated when needed and ifail is set to a positive value 
*     if any inversions were not safe (ifail=0 if they were safe)
*
*   void run_md(hmc_act_t act1,int *stat_acf,int *stat_gcr_acf,
*               int *stat_frf,int *stat_gcr_frf,int *stat_dfl_upd,int *ifail)
*     This program evolves the gauge field and the momenta on the blocks of
*     the HMC_BLOCKS grid according to the parameters set by set_hmc_parms().
*     At the end of the evolution, the program computes the active parts of
*     the HMC action and assigns the result to the elements of the structure
*     act1 (see the notes).
*      The parameters stat_gcr_* are arrays of 3 elements and stat_dfl_upd an
*     array of 2 elements. All other stat_* parameters point to a single value
*     (see the notes). The SW-term is updated when needed and ifail is set to
*     a positive value if any inversions were not safe (ifail=0 if they were
*     safe)
*
*   int accept_step(hmc_act_t act0,hmc_act_t act1,double *dact)
*     Computes the difference act1-act0 of the specified HMC actions and
*     assigns the sum of all its parts to dact (see the notes). The program  
*     returns 1 with probability min{1,exp(-dact)} and 0 otherwise. If 1 is
*     returned, the active link variables on the HMC_BLOCKS grid are swapped
*     with those in the global field. On exit the time-like link variables on
*     the global lattice at time NPROC0*L0-1 are left in the unflipped state
*     and the field is renormalized to SU(3) if it was changed
*
* Notes:
*
* The gauge field that is evolved in the course of the leap-frog integration
* is the double-precision field on the blocks. The global double-precision
* gauge field is left intact until the evolved configuration is accepted
*
* Depending on the parameters returned by hmc_parms(), the ordinary or the
* deflated SAP+GCR solver is used for the solution of the Dirac equation on
* the full lattice. The deflation subspace is generated and updated along
* the trajectory using dfl_modes() and dfl_update() [dfl/dfl_modes.c]. The
* "chronological inversion method" is silently applied depending on the 
* values of the parameters hmc.cno0, hmc.cno1 and hmc.cno_blk returned by
* hmc_parms() (see hmc/chrono.c, hmc/force1.c and hmc/force2.c)
*
* It is assumed that the parameters of the algorithm were set by calling
* set_lat_parms(), set_gcr_parms(), set_hmc_parms() and set_dfl_parms()
* (if deflation is switched on). The programs require a workspace of
*
*  max{2*nkv+1,Ns+2} (deflation is on)
*
*  2*nkv+1           (deflation is off)
*
* single-precision spinor fields, where nkv denotes the number of Krylov
* vectors returned by gcr_parms() and Ns the number of deflation modes
* returned by dfl_parms(). The number of double-precision spinor fields
* used as workspace is
*
*  8                 (deflation on,  hmc.cno0+hmc.cno1>0)
*
*  7                 (deflation on,  hmc.cno0=hmc.cno1=0)
*
*  6                 (deflation off, hmc.cno0+hmc.cno1>0)
*
*  5                 (deflation off, hmc.cno0=hmc.cno1=0)
*
* The workspace is reserved internally using the utilities in the wspace.c
* module (these must therefore also be used in the calling programs).
*
* The structures of type hmc_act_t have four elements
*
*  nb                 Number of HMC_BLOCKS blocks in the local lattice
*
*  act0[nb]           Gauge parts of the HMC action, including the action of
*                     the field momenta. The array element act0[n] is the
*                     part of the action computed on block number n
*
*  act1[nb]           Block fermion parts of the HMC action, including the 
*                     contribution of the factorized SW-term
*
*  act2[nb]           Block interaction part of the HMC action, split into
*                     block contributions that are assigned to the array
*                     elements act2[n] (act2[n]=0 if n is a white block)
*
* The total action is thus obtained by adding up all elements of act0, act1
* and act2 and by summing the result over all MPI processes. However, in
* order to avoid an excessive accumulation of rounding errors, the program
* accept_step() takes the differences of the elements of the input action
* structures *before* summing over all elements
*
* The status variables have the following meaning:
*
*  stat_rop           Average number of BiCGstab iterations that were
*                     required when calculating R_2*chi
*
*  stat_dfl_mds       Average number of GCR iterations that were required
*                     for the solution of the little Dirac equation in the
*                     course of the computation of the deflation modes
*                     (set to 0 if deflation is switched off)
*
*  stat_acf           Average number of BiCGstab iterations that were
*                     required when calculating the block action at the
*                     end of the trajectory
*
*  stat_frf           Average number of CG iterations that were required
*                     for the calculation of the block forces
*
*  stat_gcr_acf[0]    Number of SAP+GCR iterations that were required for
*                     the calculation of the action of the block-interaction
*                     term
*
*  stat_gcr_acf[1]    Average number of GCR iterations that were required
*                     for the solution of the little system in the inner
*                     cycles of the deflated SAP+GCR solver (calculation of
*                     the block-interaction action; set to 0 if deflation
*                     is switched off)
*
*  stat_gcr_acf[2]    Average number of GCR iterations that were required
*                     for the solution of the little system in the outer
*                     cycles of the deflated SAP+GCR solver (calculation of
*                     the block-interaction action; set to 0 if deflation
*                     is switched off)
*                     
*  stat_gcr_frf[0]    Average number of SAP+GCR iterations that were
*                     required for the calculation of the block-interaction
*                     force
*
*  stat_gcr_frf[1]    Same as stat_gcr_acf[1], but for the block-interaction
*                     force
*
*  stat_gcr_frf[2]    Same as stat_gcr_acf[2], but for the block-interaction
*                     force
*
*  stat_dfl_upd[0]    Total number of update steps applied to the deflation
*                     subspace in the course of the trajectory (set to 0 if
*                     deflation is switched off)
*
*  stat_dfl_upd[1]    Average number of GCR iterations that were required
*                     for the solution of the little system in the course
*                     of the updates of the deflation subspace (set to 0
*                     if deflation is switched off)
*
* In all cases, the programs return negative status values if the associated
* task could not be completed successfully
*
*******************************************************************************/

#define LPFROG_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "random.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "linalg.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "hmc.h"
#include "dfl.h"
#include "global.h"

#define DEBUG 0

static int iwd0,iwd1,nbs=0;
static double *acts0,*acts1,*acts2;

#if (defined SSE2)
#include "sse2.h"
static pauli_dble minv __attribute__ ((aligned (16)));
#else
static pauli_dble minv;
#endif


void alloc_hmc_action(hmc_act_t *act)
{
   int nb,isw;
   double *a,*am;

   blk_list(HMC_BLOCKS,&nb,&isw);
   error(nb==0,1,"alloc_hmc_action [lpfrog.c]",
         "HMC_BLOCKS grid is not allocated");
   
   a=amalloc(3*nb*sizeof(*a),3);
   error(a==NULL,1,"alloc_hmc_action [lpfrog.c]",
         "Unable to allocate action data arrays");

   (*act).nb=nb;
   (*act).act0=a;
   (*act).act1=a+nb;
   (*act).act2=a+2*nb;

   am=a+3*nb;
   
   for (;a<am;a++)
      (*a)=0.0;
}


static int apply_Qee_inv(int vol,pauli_dble *m,spinor_dble *s,spinor_dble *r)
{
   int ifail;
   weyl_dble *rs,*rr,*rsm;

   ifail=0;
   rs=(weyl_dble*)(s);
   rr=(weyl_dble*)(r);
   rsm=rs+2*vol;

   for (;rs<rsm;rs++)
   {
      ifail|=inv_pauli_dble(m,&minv);
      mul_pauli_dble(&minv,rs,rr);
      m+=1;
      rr+=1;
   }

   return ifail;
}


void start_md(hmc_act_t act0,int *stat_rop,int *stat_dfl_mds,int *ifail)
{
   int n,nb,nbh,isw,flag0,flag1,flag2,dmy0,dmy1;
   double fact;
   block_t *b;
   hmc_t *hflds;
   lat_parms_t lat;
   gcr_parms_t gcr;
   hmc_parms_t hmc;

   lat=lat_parms();
   gcr=gcr_parms();
   hmc=hmc_parms();
   
   error_root((gcr.nb==0)||(hmc.nb==0),1,"start_md [lpfrog.c]",
              "Block grids are not allocated");

   n=query_flags(BCD_FLIPPED);
   if (((hmc.bc==1)&&(n!=0))||((hmc.bc==-1)&&(n!=1)))
      flipbcd();

   assign_ud2udbgr(HMC_BLOCKS);
   swb_term();

   if (hmc.dfl==1)
      dfl_modes(DEBUG>=2,stat_dfl_mds,&flag0);
   else
   {
      (*stat_dfl_mds)=0;
      flag0=0;
   }
   
   reset_chrono_blk();
   reset_chrono();
   
   random_chi();
   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();
   nbh=nb/2;
   fact=lat.beta/6.0;
   flag1=0;
   
   for (n=0;n<nb;n++)
   {   
      random_alg((*hflds).nlks,(*hflds).mom);
      act0.act0[n]=0.5*norm_square_alg((*hflds).nlks,(*hflds).mom);
      act0.act0[n]+=fact*hmc_acg(n);

      act0.act1[n]=hmc_acf(n,0,1,&dmy0,&dmy1);
      random_sd((*b).volh,(*b).sd[0],1.0);
      act0.act1[n]+=norm_square_dble((*b).volh,0,(*b).sd[0]);      
      flag1|=invert_blk_swd(HMC_BLOCKS,n,ODD_PTS);
      Qhat_blk_dble(b,0,1);
      flag1|=apply_Qee_inv((*b).volh,(*b).swd,(*b).sd[1],(*hflds).phi);

      if (((isw==0)&&(n<nbh))||((isw==1)&&(n>=nbh)))
         act0.act2[n]=norm_square_dble((*hflds).nbnd,0,(*hflds).chi);
      else
         act0.act2[n]=0.0;

      b+=1;
      hflds+=1;
   }
   
   ROp(stat_rop,&flag2);

   flag0|=(flag1|flag2);
   MPI_Reduce(&flag0,ifail,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(ifail,1,MPI_INT,0,MPI_COMM_WORLD);
   
   error_chk();   
}


static void updt_mom(int n,double eps)
{
   su3_alg_dble *frc,*frm,*mom;
   hmc_t *hflds;

   hflds=hmcflds()+n;
   mom=(*hflds).mom;
   frc=(*hflds).frc;
   frm=frc+(*hflds).nlks;

   for (;frc<frm;frc++)
   {
      (*mom).c1-=eps*(*frc).c1;
      (*mom).c2-=eps*(*frc).c2;      
      (*mom).c3-=eps*(*frc).c3;
      (*mom).c4-=eps*(*frc).c4;
      (*mom).c5-=eps*(*frc).c5;
      (*mom).c6-=eps*(*frc).c6;      
      (*mom).c7-=eps*(*frc).c7;
      (*mom).c8-=eps*(*frc).c8;

      mom+=1;
   }
}


static void T_U(int n,double eps)
{
   int nb,isw,*lk,*lkm;
   su3_dble *ub,*u;
   su3_alg_dble *mom;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;
   hflds=hmcflds()+n;
   
   ub=(*b).ud[0];
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   mom=(*hflds).mom;
   
   for (;lk<lkm;lk++)
   {
      u=ub+(*lk);
      ExpX(eps,mom,u);
      mom+=1;
   }

   set_blk_flags(HMC_BLOCKS,NEW_ACTIVE_UD,n);
}


static void T_G(int n,double eps)
{
   hmc_frg(n);
   updt_mom(n,eps);
}


static void T_F(int n,double eps,int *status,int *ifail)
{
   int stat,flag;

   hmc_frf(n,&stat,&flag);
   
   if (stat<0)
      (*status)=stat;
   else
      (*status)+=stat;

   (*ifail)|=flag;

   if (((*status)<0)||((*ifail)!=0))
      return;

   updt_mom(n,eps);
}


static void T_GCR(double eps,int nmx_dfl,int *stat_dfl,int *status,int *ifail)
{
   int n,nb,isw;
   int stat0,stat1[3],flag;

   hmc_gcr_vec(iwd0,iwd1,nmx_dfl,&stat0,stat1,&flag);

#if (DEBUG>=1)
   message("hmc_gcr_vec: nmx_dfl = %d, stat_dfl = %d, "
           "<status> = %d (little: %d, %d), ifail = %d\n",
           nmx_dfl,stat0,stat1[0]/2,stat1[1]/2,stat1[2]/2,flag);
#endif

   if ((*stat_dfl)>=0)
   {
      if (stat0<0)
         (*stat_dfl)=stat0;
      else
         (*stat_dfl)+=stat0;
   }

   for (n=0;n<3;n++)
   {
      if (status[n]>=0)
      {
         if (stat1[n]<0)
            status[n]=stat1[n];
         else
            status[n]+=stat1[n];
      }
   }
   
   (*ifail)|=flag;

   if (((*stat_dfl)<0)||(status[0]<0)||(status[2]<0)||((*ifail)!=0))
      return;
   
   blk_list(HMC_BLOCKS,&nb,&isw);
   
   for (n=0;n<nb;n++)
   {
      hmc_gcr_frf(n,iwd0,iwd1);
      updt_mom(n,eps);
   }
}


void run_md(hmc_act_t act1,int *stat_acf,int *stat_gcr_acf,
            int *stat_frf,int *stat_gcr_frf,int *stat_dfl_upd,int *ifail)
{
   int n,nh,nb,cno;
   int n0,n1,n2;
   int k0,k1,k2;
   int stat,flag,stat_min,stat_sum,dmy;
   int stat_ref,stat_new,stat_old,stat_del,stat_inc;
   double fact,eps0,eps1,eps2;
   spinor_dble **wsd0,**wsd1;
   hmc_t *hflds;
   lat_parms_t lat;
   hmc_parms_t hmc;
   dfl_parms_t dfl;   

   wsd0=reserve_wsd(1);
   wsd1=reserve_wsd(1);
   iwd0=(wsd0[0]-psd[0][0])/NSPIN;
   iwd1=(wsd1[0]-psd[0][0])/NSPIN;   

   lat=lat_parms();
   hmc=hmc_parms();

   nb=hmc.nb;
   n0=hmc.N0;
   n1=hmc.N1;
   n2=hmc.N2;

   fact=lat.beta/6.0;   
   eps2=hmc.tau/(double)(n2);
   eps1=eps2/(double)(n1);   
   eps0=eps1/(double)(n0);

   for (n=0;n<nb;n++)
   {
      act1.act0[n]=0.0;
      act1.act1[n]=0.0;
      act1.act2[n]=0.0;
   }
   
   (*stat_acf)=0;
   (*stat_frf)=0;

   stat_gcr_acf[0]=0;
   stat_gcr_acf[1]=0;
   stat_gcr_acf[2]=0;   

   stat_gcr_frf[0]=0;
   stat_gcr_frf[1]=0;
   stat_gcr_frf[2]=0;   

   stat_dfl_upd[0]=0;
   stat_dfl_upd[1]=0;
   
   (*ifail)=0;
   stat_ref=0;
   stat_new=0;
   stat_old=0;
   stat_inc=0;
   stat_del=0;
   dmy=0;

   if (hmc.dfl)
   {
      dfl=dfl_parms();
      stat_del=dfl.Ns/2;
   }

   if (hmc.cno0>hmc.cno1)
      cno=hmc.cno0;
   else
      cno=hmc.cno1;
   
   for (k2=1;k2<=n2;k2++)
   {
      if (k2==1)
      {
         T_GCR(0.5*eps2,0,&dmy,stat_gcr_frf,ifail);

         if (hmc.dfl)
         {
            stat_ref=stat_gcr_frf[0];
            stat_old=stat_ref;
         }
      }
      else
      {
         update_swdbgr2swd();

         if ((hmc.dfl)&&(stat_inc>=stat_del))
         {
            T_GCR(eps2,1,stat_dfl_upd+1,stat_gcr_frf,ifail);               
            stat_dfl_upd[0]+=1;
            stat_inc=0;
         }
         else
            T_GCR(eps2,0,&dmy,stat_gcr_frf,ifail);

         if (hmc.dfl)
         {
            stat_new=stat_gcr_frf[0]-stat_old;
            stat_old=stat_gcr_frf[0];

            if (stat_new<stat_ref)
               stat_ref=stat_new;
            else
               stat_inc+=(stat_new-stat_ref);

            if (k2<cno)
               stat_inc+=((stat_new+8)/16);
         }
      }

      if ((stat_dfl_upd[1]<0)||(stat_gcr_frf[0]<0)||(stat_gcr_frf[2]<0)||
          ((*ifail)!=0))
         break;

      stat=0;
      flag=0;
      
      for (n=0;n<nb;n++)
      {
         for (k1=1;k1<=n1;k1++)
         {
            if ((k2==1)&&(k1==1))
               T_F(n,0.5*eps1,&stat,&flag);
            else
               T_F(n,eps1,&stat,&flag);

            if ((stat<0)||(flag!=0))
               break;
            
            for (k0=1;k0<=n0;k0++)
            {
               if ((k2==1)&&(k1==1)&&(k0==1))
                  T_G(n,0.5*fact*eps0);
               else
                  T_G(n,fact*eps0); 

               T_U(n,eps0);

               if ((k2==n2)&&(k1==n1)&&(k0==n0))
                  T_G(n,0.5*fact*eps0); 
            }
         
            if ((k2==n2)&&(k1==n1))
               T_F(n,0.5*eps1,&stat,&flag);

            if ((stat<0)||(flag!=0))
               break;
         }

         if ((stat<0)||(flag!=0))
            break;
      }

      MPI_Reduce(&stat,&stat_min,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
      MPI_Reduce(&stat,&stat_sum,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Reduce(&flag,ifail,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
      MPI_Bcast(&stat_min,1,MPI_INT,0,MPI_COMM_WORLD);      
      MPI_Bcast(&stat_sum,1,MPI_INT,0,MPI_COMM_WORLD);            
      MPI_Bcast(ifail,1,MPI_INT,0,MPI_COMM_WORLD);

      if ((*stat_frf)>=0)
      {
         if (stat_min<0)
            (*stat_frf)=stat_min;
         else
            (*stat_frf)+=stat_sum;
      }
      
      if (((*stat_frf)<0)||((*ifail)!=0))
      {
         error_chk();
         break;
      }
      
      if (k2==n2)
      {
         update_swdbgr2swd();

         if ((hmc.dfl)&&(stat_inc>=stat_del))
         {
            T_GCR(0.5*eps2,1,stat_dfl_upd+1,stat_gcr_frf,ifail);
            stat_dfl_upd[0]+=1;
         }
         else
            T_GCR(0.5*eps2,0,&dmy,stat_gcr_frf,ifail);
      }
   }

   release_wsd();
   
   if ((*stat_frf)>0)
   {
      n=NPROC*nb*(n1*n2+1);
      nh=n/2;
      (*stat_frf)=((*stat_frf)+nh)/n;
   }

   if (stat_dfl_upd[1]>0)
   {
      n=stat_dfl_upd[0];
      nh=n/2;
      stat_dfl_upd[1]=(stat_dfl_upd[1]+nh)/n;
   }

   n=2*(n2+1);
   nh=n/2;
   if (stat_gcr_frf[0]>0)
      stat_gcr_frf[0]=(stat_gcr_frf[0]+nh)/n;
   if (stat_gcr_frf[1]>0)
      stat_gcr_frf[1]=(stat_gcr_frf[1]+nh)/n;      
   if (stat_gcr_frf[2]>0)
      stat_gcr_frf[2]=(stat_gcr_frf[2]+nh)/n;      
   
   if (((*stat_frf)<0)||(stat_dfl_upd[1]<0)||(stat_gcr_frf[0]<0)||
       (stat_gcr_frf[2]<0)||((*ifail)!=0))
   {
      release_wsd();
      return;
   }

   hmc_gcr_acf(iwd0,0,act1.act2,&dmy,stat_gcr_acf,ifail);

#if (DEBUG>=1)   
   message("hmc_gcr_acf: stat_gcr_acf = %d (little: %d, %d), ifail = %d\n",
           stat_gcr_acf[0],stat_gcr_acf[1],stat_gcr_acf[2],(*ifail));
#endif
   
   release_wsd();
   
   if ((stat_gcr_acf[0]<0)||(stat_gcr_acf[2]<0)||((*ifail)!=0))
      return;

   hflds=hmcflds();
   stat=0;
   flag=0;
   
   for (n=0;n<nb;n++)
   {
      act1.act0[n]=0.5*norm_square_alg((*hflds).nlks,(*hflds).mom);
      act1.act0[n]+=fact*hmc_acg(n);

      act1.act1[n]=hmc_acf(n,1,1,stat_acf,&flag);

      if ((*stat_acf)>=0)
         stat+=(*stat_acf);
      else
         stat=(*stat_acf);

      if ((stat<0)||(flag!=0))
         break;
      
      hflds+=1;
   }

   MPI_Reduce(&stat,&stat_min,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);
   MPI_Reduce(&stat,&stat_sum,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Reduce(&flag,ifail,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(&stat_min,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(&stat_sum,1,MPI_INT,0,MPI_COMM_WORLD);            
   MPI_Bcast(ifail,1,MPI_INT,0,MPI_COMM_WORLD);
      
   if (stat_min<0)
      (*stat_acf)=stat_min;
   else
   {
      n=NPROC*nb;
      nh=n/2;
      (*stat_acf)=(stat_sum+nh)/n;
   }
   
   error_chk();   
}


static void alloc_acts(int nb)
{
   if (nbs>0)
      afree(acts0);

   if (nb>0)
   {
      acts0=amalloc(3*nb*sizeof(*acts0),3);
      error(acts0==NULL,1,"alloc_acts [lpfrog.c]",
            "Unable to allocate auxiliary arrays");

      acts1=acts0+nb;
      acts2=acts1+nb;
      nbs=nb;
   }
   else
   {
      acts0=NULL;
      acts1=NULL;
      acts2=NULL;
      nbs=0;
   }
}


int accept_step(hmc_act_t act0,hmc_act_t act1,double *dact)
{
   int my_rank,iac;
   int nb,n;
   double sm,r;

   nb=act0.nb;

   if (nb!=nbs)
      alloc_acts(nb);

   for (n=0;n<nb;n++)
   {
      acts0[n]=act1.act0[n]-act0.act0[n];
      acts1[n]=act1.act1[n]-act0.act1[n];
      acts2[n]=act1.act2[n]-act0.act2[n];
   }

   sm=0.0;
   
   for (n=0;n<nb;n++)
      sm+=(acts0[n]+acts1[n]+acts2[n]);

   MPI_Reduce(&sm,dact,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
   MPI_Bcast(dact,1,MPI_DOUBLE,0,MPI_COMM_WORLD);     
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   
   if (my_rank==0)
   {
      ranlxd(&r,1);

      if ((*dact)<=0.0)
         iac=1;
      else if (r<=exp(-(*dact)))
         iac=1;
      else
         iac=0;
   }
   
   MPI_Bcast(&iac,1,MPI_INT,0,MPI_COMM_WORLD);   

   if (iac==1)
      swap_active(HMC_BLOCKS);

   if (query_flags(BCD_FLIPPED)!=0)
      flipbcd();

   if (iac==1)
      renormalize_ud();
   
   return iac;
}

