
/*******************************************************************************
*
* File random_su2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generatation of random SU(2) matrices
*
* The externally accessible function is
*
*   void random_su2(float rho,float s[])
*     Computes a random vector s[4] with probability density proportional
*     to exp(rho*s[0])*delta(1-s^2) assuming rho>=0
*
*******************************************************************************/

#define RANDOM_SU2_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "start.h"
#include "random.h"

#if (defined SSE)

#include "sse.h"
#include "misc.h"

#define NVEC (32)
#define NRAN (2*NVEC)

static int i_vec=NVEC,i_y=NRAN,i_v=NRAN;
static float vec1[NVEC],vec2[NVEC],vec3[NVEC] __attribute__ ((aligned (16)));
static float r[NRAN],u[NRAN],v[NRAN],y[NRAN] __attribute__ ((aligned (16)));
static sse_float xvec,yvec,zvec;
static sse_float xfm1={-1.0f,-1.0f,-1.0f,-1.0f};
static sse_float xfh={0.5f,0.5f,0.5f,0.5f};
static sse_float xf1={1.0f,1.0f,1.0f,1.0f};
static sse_float xf2={2.0f,2.0f,2.0f,2.0f};


static void update_vec(void)
{
   int i;

   ranlxs(r,NRAN);

   for (i=0;i<NVEC;i+=4)
   {
      __asm__ __volatile__ ("movaps %1, %%xmm0 \n\t"
                            "mulps %2, %%xmm0 \n\t"
                            "subps %3, %%xmm0 \n\t"
                            "movaps %%xmm0, %0"
                            :
                            "=m" (xvec)
                            :
                            "m" (r[NVEC+i]),
                            "m" (xf2),
                            "m" (xf1)
                            :
                            "xmm0");
      
      sse_sincos(&xvec,&yvec,&zvec);

      __asm__ __volatile__ ("movaps %3, %%xmm1 \n\t"
                            "movaps %4, %%xmm2 \n\t"
                            "movaps %5, %%xmm3 \n\t"
                            "mulps %6, %%xmm3 \n\t"
                            "subps %7, %%xmm3 \n\t"
                            "movaps %%xmm3, %0 \n\t"
                            "movaps %7, %%xmm4 \n\t"
                            "mulps %%xmm3, %%xmm3 \n\t"
                            "subps %%xmm3, %%xmm4 \n\t"
                            "sqrtps %%xmm4, %%xmm5 \n\t"
                            "mulps %%xmm5, %%xmm1 \n\t"
                            "mulps %%xmm5, %%xmm2 \n\t"
                            "movaps %%xmm1, %1 \n\t"
                            "movaps %%xmm2, %2"
                            :
                            "=m" (vec1[i]),
                            "=m" (vec2[i]),
                            "=m" (vec3[i])
                            :
                            "m" (yvec),
                            "m" (zvec),
                            "m" (r[i]),
                            "m" (xf2),
                            "m" (xf1)
                            :
                            "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");
   }

   i_vec=0;
}


static void update_y(void)
{
   int i;

   ranlxs(y,NRAN);
   ranlxs(u,NRAN);
   ranlxs(r,NRAN);

   for (i=0;i<NVEC;i+=4)
   {
      __asm__ __volatile__ ("movaps %1, %%xmm0 \n\t"
                            "mulps %2, %%xmm0 \n\t"
                            "movaps %%xmm0, %0"
                            :
                            "=m" (xvec)
                            :
                            "m" (y[i]),
                            "m" (xfm1)
                            :
                            "xmm0");
      
      sse_ln1px(&xvec,&yvec);
      
      __asm__ __volatile__ ("movaps %1, %%xmm1 \n\t"
                            "mulps %2, %%xmm1 \n\t"
                            "movaps %%xmm1, %0"
                            :
                            "=m" (y[i])
                            :
                            "m" (yvec),
                            "m" (xfm1)
                            :
                            "xmm1");
   }

   for (i=0;i<NRAN;i+=4)
   {
      __asm__ __volatile__ ("movaps %1, %%xmm0 \n\t"
                            "mulps %2, %%xmm0 \n\t"
                            "movaps %%xmm0, %0"
                            :
                            "=m" (xvec)
                            :
                            "m" (u[i]),
                            "m" (xfm1)
                            :
                            "xmm0");
      
      sse_ln1px(&xvec,(sse_float*)(&u[i]));
   }

   for (i=0;i<NVEC;i+=4)
   {
      __asm__ __volatile__ ("movaps %1, %%xmm0 \n\t"
                            "mulps %2, %%xmm0 \n\t"
                            "movaps %%xmm0, %0"
                            :
                            "=m" (xvec)
                            :
                            "m" (y[NVEC+i]),
                            "m" (xfh)
                            :
                            "xmm0");

      sse_sincos(&xvec,&yvec,&zvec);

      __asm__ __volatile__ ("movaps %2, %%xmm1 \n\t"
                            "movaps %3, %%xmm2 \n\t"
                            "movaps %4, %%xmm3 \n\t"
                            "movaps %5, %%xmm4 \n\t"
                            "movaps %6, %%xmm5 \n\t"
                            "mulps %%xmm1, %%xmm1 \n\t"
                            "mulps %%xmm2, %%xmm2 \n\t"
                            "mulps %%xmm3, %%xmm1 \n\t"
                            "mulps %%xmm3, %%xmm2 \n\t"
                            "subps %%xmm4, %%xmm1 \n\t"
                            "subps %%xmm5, %%xmm2 \n\t"
                            "movaps %%xmm1, %0 \n\t"
                            "movaps %%xmm2, %1"
                            :
                            "=m" (y[i]),
                            "=m" (y[NVEC+i])
                            :
                            "m" (yvec),
                            "m" (zvec),
                            "m" (y[i]),
                            "m" (u[i]),
                            "m" (u[NVEC+i])
                            :
                            "xmm1", "xmm2", "xmm3", "xmm4","xmm5");

      __asm__ __volatile__ ("movaps %2, %%xmm0 \n\t"
                            "movaps %3, %%xmm1 \n\t"
                            "mulps %%xmm0,%%xmm0 \n\t"
                            "mulps %%xmm1,%%xmm1 \n\t"
                            "addps %%xmm0, %%xmm0 \n\t"
                            "addps %%xmm1, %%xmm1 \n\t"
                            "movaps %%xmm0, %0 \n\t"
                            "movaps %%xmm1, %1"
                            :
                            "=m" (u[i]),
                            "=m" (u[NVEC+i])
                            :
                            "m" (r[i]),
                            "m" (r[NVEC+i])
                            :
                            "xmm0", "xmm1");
   }

   i_y=0;
}

#else

#define NVEC (32)
#define NRAN (2*NVEC)
#define PI_HALF 1.5707964f
#define PI 3.1415927f
#define TWO_PI 6.2831854f

static int i_vec=NVEC,i_y=NRAN,i_v=NRAN;
static float vec1[NVEC],vec2[NVEC],vec3[NVEC];
static float r[NRAN],u[NRAN],v[NRAN],y[NRAN];


static void update_vec(void)
{
   int i;
   float r1,r2,rsq;

   ranlxs(r,NRAN);

   for (i=0;i<NVEC;i++)
   {
      r1=2.0f*r[i]-1.0f;
      r2=TWO_PI*r[NVEC+i]-PI;
      rsq=(float)(sqrt((double)(1.0f-r1*r1)));

      vec1[i]=r1;
      vec2[i]=rsq*sin(r2);
      vec3[i]=rsq*cos(r2);
   }

   i_vec=0;
}


static void update_y(void)
{
   int i;
   float r1,r2,r3,r4,s,c;

   ranlxs(y,NRAN);
   ranlxs(u,NRAN);
   ranlxs(r,NRAN);

   for (i=0;i<NVEC;i++)
   {
      r1=-(float)(log((double)(1.0f-y[i])));
      r2=PI_HALF*y[NVEC+i];
      r3=(float)(log((double)(1.0f-u[i])));
      r4=(float)(log((double)(1.0f-u[NVEC+i])));

      s=sin(r2);
      s*=s;
      c=1.0f-s;

      y[i]=r1*s-r3;
      y[NVEC+i]=r1*c-r4;

      r1=r[i]*r[i];
      r2=r[NVEC+i]*r[NVEC+i];
      u[i]=r1+r1;
      u[NVEC+i]=r2+r2;
   }

   i_y=0;
}

#endif

void random_su2(float rho,float s[])
{
   float rhoinv,s0p1,ut,rt;
   float s0,s1,s2,s3,sq;

   if (i_vec==NVEC)
      update_vec();

   if (rho>1.5f)
   {
      rhoinv=1.0f/rho;

      for (;;)
      {
         if (i_y==NRAN)
            update_y();

         s0p1=2.0f-rhoinv*y[i_y];
         ut=u[i_y++];

         if (ut<=s0p1)
            break;
      }
   }
   else if (rho>0.3f)
   {
      rhoinv=1.0f/rho;
      rt=(float)(exp((double)(rho+rho))-1.0);

      for (;;)
      {
         if (i_v==NRAN)
         {
            ranlxs(v,NRAN);
            i_v=0;
         }

         s0p1=rhoinv*(float)(log(1.0+(double)(rt*v[i_v++])));
         ut=v[i_v++];

         if ((ut*ut)<=(s0p1*(2.0f-s0p1)))
            break;
      }
   }
   else
   {
      for (;;)
      {
         if (i_v==NRAN)
         {
            ranlxs(v,NRAN);
            i_v=0;
         }

         s0p1=2.0f*v[i_v++];
         rt=(float)(exp((double)(rho*(s0p1-2.0f))));
         ut=v[i_v++];

         if ((ut*ut)<=(s0p1*(2.0f-s0p1)*rt*rt))
            break;
      }
   }

   sq=(float)(sqrt((double)(s0p1*(2.0f-s0p1))));
   s0=s0p1-1.0f;
   s1=sq*vec1[i_vec];
   s2=sq*vec2[i_vec];
   s3=sq*vec3[i_vec];

   sq=s0*s0+s1*s1+s2*s2+s3*s3;
   sq=1.5f-0.5f*sq;

   s[0]=sq*s0;
   s[1]=sq*s1;
   s[2]=sq*s2;
   s[3]=sq*s3;

   i_vec+=1;
}

